/**
 * vim: set ts=4 :
 * =============================================================================
 * SourceMod Client Preferences Extension
 * Copyright (C) 2004-2008 AlliedModders LLC.  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, AlliedModders LLC gives you permission to link the
 * code of this program (as well as its derivative works) to "Half-Life 2," the
 * "Source Engine," the "SourcePawn JIT," and any Game MODs that run on software
 * by the Valve Corporation.  You must obey the GNU General Public License in
 * all respects for all other code used.  Additionally, AlliedModders LLC grants
 * this exception to all derivative works.  AlliedModders LLC defines further
 * exceptions, found in LICENSE.txt (as of this writing, version JULY-31-2007),
 * or <http://www.sourcemod.net/license.php>.
 *
 * Version: $Id$
 */

#include <sourcemod_version.h>
#include "extension.h"

/**
 * @file extension.cpp
 * @brief Implement extension code here.
 */

ClientPrefs g_ClientPrefs;		/**< Global singleton for extension's main interface */

SMEXT_LINK(&g_ClientPrefs);

HandleType_t g_CookieType = 0;
CookieTypeHandler g_CookieTypeHandler;

HandleType_t g_CookieIterator = 0;
CookieIteratorHandler g_CookieIteratorHandler;
DbDriver g_DriverType;

bool ClientPrefs::SDK_OnLoad(char *error, size_t maxlength, bool late)
{
	queryMutex = threader->MakeMutex();
	cookieMutex = threader->MakeMutex();

	DBInfo = dbi->FindDatabaseConf("clientprefs");

	if (DBInfo == NULL)
	{
		DBInfo = dbi->FindDatabaseConf("default");

		if (DBInfo == NULL)
		{
			DBInfo = dbi->FindDatabaseConf("storage-local");
		}
	}
	
	if (DBInfo == NULL)
	{
		snprintf(error, maxlength, "Could not find any suitable database configs");
		return false;
	}

	if (DBInfo->driver && DBInfo->driver[0] != '\0')
	{
		Driver = dbi->FindOrLoadDriver(DBInfo->driver);
	}
	else
	{
		Driver = dbi->GetDefaultDriver();
	}

	if (Driver == NULL)
	{
		snprintf(error, maxlength, "Could not load DB Driver \"%s\"", DBInfo->driver);
		return false;
	}

	databaseLoading = true;
	TQueryOp *op = new TQueryOp(Query_Connect, 0);
	dbi->AddToThreadQueue(op, PrioQueue_High);

	dbi->AddDependency(myself, Driver);

	sharesys->AddNatives(myself, g_ClientPrefNatives);
	sharesys->RegisterLibrary(myself, "clientprefs");
	identity = sharesys->CreateIdentity(sharesys->CreateIdentType("ClientPrefs"), this);
	g_CookieManager.cookieDataLoadedForward = forwards->CreateForward("OnClientCookiesCached", ET_Ignore, 1, NULL, Param_Cell);

	g_CookieType = handlesys->CreateType("Cookie", 
		&g_CookieTypeHandler, 
		0, 
		NULL, 
		NULL, 
		myself->GetIdentity(), 
		NULL);

	g_CookieIterator = handlesys->CreateType("CookieIterator", 
		&g_CookieIteratorHandler, 
		0, 
		NULL, 
		NULL, 
		myself->GetIdentity(), 
		NULL);

	IMenuStyle *style = menus->GetDefaultStyle();
	g_CookieManager.clientMenu = style->CreateMenu(&g_Handler, identity);
	g_CookieManager.clientMenu->SetDefaultTitle("Client Settings:");

	plsys->AddPluginsListener(&g_CookieManager);

	phrases = translator->CreatePhraseCollection();
	phrases->AddPhraseFile("clientprefs.phrases");
	phrases->AddPhraseFile("common.phrases");

	if (late)
	{
		this->CatchLateLoadClients();
	}

	return true;
}

void ClientPrefs::SDK_OnAllLoaded()
{
	playerhelpers->AddClientListener(&g_CookieManager);
}

bool ClientPrefs::QueryInterfaceDrop(SMInterface *pInterface)
{
	if ((void *)pInterface == (void *)(Database->GetDriver()))
	{
		return false;
	}

	return true;
}

void ClientPrefs::NotifyInterfaceDrop(SMInterface *pInterface)
{
	if (Database != NULL && (void *)pInterface == (void *)(Database->GetDriver()))
	{
		Database->Close();
		Database = NULL;
	}
}

void ClientPrefs::SDK_OnUnload()
{
	handlesys->RemoveType(g_CookieType, myself->GetIdentity());
	handlesys->RemoveType(g_CookieIterator, myself->GetIdentity());

	g_CookieManager.Unload();

	if (Database != NULL)
	{
		Database->Close();
		Database = NULL;
	}

	if (g_CookieManager.cookieDataLoadedForward != NULL)
	{
		forwards->ReleaseForward(g_CookieManager.cookieDataLoadedForward);
		g_CookieManager.cookieDataLoadedForward = NULL;
	}

	if (g_CookieManager.clientMenu != NULL)
	{
		Handle_t menuHandle = g_CookieManager.clientMenu->GetHandle();
		
		if (menuHandle != BAD_HANDLE)
		{
			HandleSecurity sec = HandleSecurity(identity, identity);
			HandleError err = handlesys->FreeHandle(menuHandle, &sec);
			if (HandleError_None != err)
			{
				g_pSM->LogError(myself, "Error %d when attempting to free client menu handle", err);
			}
		}
		
		g_CookieManager.clientMenu = NULL;
	}

	if (phrases != NULL)
	{
		phrases->Destroy();
		phrases = NULL;
	}

	plsys->RemovePluginsListener(&g_CookieManager);
	playerhelpers->RemoveClientListener(&g_CookieManager);

	queryMutex->DestroyThis();
	cookieMutex->DestroyThis();
}

void ClientPrefs::OnCoreMapStart(edict_t *pEdictList, int edictCount, int clientMax)
{
	this->AttemptReconnection();
}

void ClientPrefs::AttemptReconnection()
{
	if (Database || databaseLoading)
	{
		return; /* We're already loading, or have loaded. */
	}
	
	g_pSM->LogMessage(myself, "Attempting to reconnect to database...");
	databaseLoading = true;
	
	TQueryOp *op = new TQueryOp(Query_Connect, 0);
	dbi->AddToThreadQueue(op, PrioQueue_High);
	
	this->CatchLateLoadClients(); /* DB reconnection, we should check if we missed anyone... */
}

void ClientPrefs::DatabaseConnect()
{
	char error[256];
	int errCode = 0;

	Database = Driver->Connect(DBInfo, true, error, sizeof(error));

	if (Database == NULL)
	{
		g_pSM->LogError(myself, error);
		databaseLoading = false;
		return;
	}

	const char *identifier = Driver->GetIdentifier();

	if (strcmp(identifier, "sqlite") == 0)
	{
		g_DriverType = Driver_SQLite;

		if (!Database->DoSimpleQuery(
				"CREATE TABLE IF NOT EXISTS sm_cookies  \
				( \
					id INTEGER PRIMARY KEY AUTOINCREMENT, \
					name varchar(30) NOT NULL UNIQUE, \
					description varchar(255), \
					access INTEGER \
				)"))
		{
			g_pSM->LogMessage(myself, "Failed to CreateTable sm_cookies: %s", Database->GetError());
			goto fatal_fail;
		}

		if (!Database->DoSimpleQuery(
				"CREATE TABLE IF NOT EXISTS sm_cookie_cache \
				( \
					player varchar(65) NOT NULL, \
					cookie_id int(10) NOT NULL, \
					value varchar(100), \
					timestamp int, \
					PRIMARY KEY (player, cookie_id) \
				)"))
		{
			g_pSM->LogMessage(myself, "Failed to CreateTable sm_cookie_cache: %s", Database->GetError());
			goto fatal_fail;
		}
	}
	else if (strcmp(identifier, "mysql") == 0)
	{
		g_DriverType = Driver_MySQL;

		if (!Database->DoSimpleQuery(
				"CREATE TABLE IF NOT EXISTS sm_cookies \
				( \
					id INTEGER unsigned NOT NULL auto_increment, \
					name varchar(30) NOT NULL UNIQUE, \
					description varchar(255), \
					access INTEGER, \
					PRIMARY KEY (id) \
				)"))
		{
			g_pSM->LogMessage(myself, "Failed to CreateTable sm_cookies: %s", Database->GetError());
			goto fatal_fail;
		}

		if (!Database->DoSimpleQuery(
				"CREATE TABLE IF NOT EXISTS sm_cookie_cache \
				( \
					player varchar(65) NOT NULL, \
					cookie_id int(10) NOT NULL, \
					value varchar(100), \
					timestamp int NOT NULL, \
					PRIMARY KEY (player, cookie_id) \
				)"))
		{
			g_pSM->LogMessage(myself, "Failed to CreateTable sm_cookie_cache: %s", Database->GetError());
			goto fatal_fail;
		}
	}
	else
	{
		g_pSM->LogError(myself, "Unsupported driver \"%s\"", identifier);
		goto fatal_fail;
	}

	databaseLoading = false;

	this->ProcessQueryCache();	
	return;

fatal_fail:
	Database->Close();
	Database = NULL;
	databaseLoading = false;
}

bool ClientPrefs::AddQueryToQueue( TQueryOp *query )
{
	queryMutex->Lock();
	if (Database == NULL)
	{
		cachedQueries.push_back(query);
		queryMutex->Unlock();
		return false;
	}
	
	if (!cachedQueries.empty())
	{
		queryMutex->Unlock();
		this->ProcessQueryCache();
	}
	else
	{
		queryMutex->Unlock();
	}

	query->SetDatabase(Database);
	dbi->AddToThreadQueue(query, PrioQueue_Normal);
	return true;
}

void ClientPrefs::ProcessQueryCache()
{
	if (Database == NULL)
	{
		return;
	}

	queryMutex->Lock();
	TQueryOp *op;
	for (SourceHook::List<TQueryOp *>::iterator iter = cachedQueries.begin(); iter != cachedQueries.end(); iter++)
	{
		op = *iter;
		op->SetDatabase(Database);
		dbi->AddToThreadQueue(op, PrioQueue_Normal);
	}

	cachedQueries.clear();
	queryMutex->Unlock();
}

size_t IsAuthIdConnected(char *authID)
{
	IGamePlayer *player;
	const char *authString;
	
	for (int playerIndex = playerhelpers->GetMaxClients()+1; --playerIndex > 0;)
	{
		player = playerhelpers->GetGamePlayer(playerIndex);
		if (player == NULL || !player->IsConnected())
		{
			continue;
		}
		
		authString = player->GetAuthString();
		
		if (authString == NULL || authString[0] == '\0')
		{
			continue;
		}

		if (strcmp(authString, authID) == 0)
		{
			return playerIndex;
		}
	}
	return 0;
}

void ClientPrefs::CatchLateLoadClients()
{
	IGamePlayer *pPlayer;
	for (int i = playerhelpers->GetMaxClients()+1; --i > 0;)
	{
		if (g_CookieManager.AreClientCookiesPening(i) || g_CookieManager.AreClientCookiesCached(i))
		{
			continue;
		}
	
		pPlayer = playerhelpers->GetGamePlayer(i);

		if (!pPlayer || !pPlayer->IsAuthorized())
		{
			continue;
		}

		g_CookieManager.OnClientAuthorized(i, pPlayer->GetAuthString());
	}
}

void ClientPrefs::ClearQueryCache(int serial)
{
	queryMutex->Lock();

	for (SourceHook::List<TQueryOp *>::iterator iter = cachedQueries.begin(); iter != cachedQueries.end();)
	{
		TQueryOp *op = *iter;
		if (op && op->PullQueryType() == Query_SelectData && op->PullQuerySerial() == serial)
 		{
			op->Destroy();
			iter = cachedQueries.erase(iter);
 		}
 		else
 		{
			iter++;
 		}
 	}
	queryMutex->Unlock();
}

bool Translate(char *buffer, 
				   size_t maxlength,
				   const char *format,
				   unsigned int numparams,
				   size_t *pOutLength,
				   ...)
{
	va_list ap;
	unsigned int i;
	const char *fail_phrase;
	void *params[MAX_TRANSLATE_PARAMS];

	if (numparams > MAX_TRANSLATE_PARAMS)
	{
		assert(false);
		return false;
	}

	va_start(ap, pOutLength);
	for (i = 0; i < numparams; i++)
	{
		params[i] = va_arg(ap, void *);
	}
	va_end(ap);

	if (!g_ClientPrefs.phrases->FormatString(buffer,
		maxlength, 
		format, 
		params,
		numparams,
		pOutLength,
		&fail_phrase))
	{
		if (fail_phrase != NULL)
		{
			g_pSM->LogError(myself, "[SM] Could not find core phrase: %s", fail_phrase);
		}
		else
		{
			g_pSM->LogError(myself, "[SM] Unknown fatal error while translating a core phrase.");
		}

		return false;
	}

	return true;
}

char * UTIL_strncpy(char * destination, const char * source, size_t num)
{
	if (source == NULL)
	{
		destination[0] = '\0';
		return destination;
	}
	
	size_t req = strlen(source);
	if (!req)
	{
		destination[0] = '\0';
		return destination;
	}
	else if (req >= num)
	{
		req = num-1;
	}
	
	strncpy(destination, source, req);
	destination[req] = '\0';
	return destination;
}

IdentityToken_t *ClientPrefs::GetIdentity() const
{
	return identity;
}

const char *ClientPrefs::GetExtensionVerString()
{
	return SM_VERSION_STRING;
}

const char *ClientPrefs::GetExtensionDateString()
{
	return SM_BUILD_TIMESTAMP;
}

ClientPrefs::ClientPrefs()
{
	Driver = NULL;
	Database = NULL;
	databaseLoading = false;
	phrases = NULL;
	DBInfo = NULL;

	cookieMutex = NULL;
	queryMutex = NULL;
	identity = NULL;
}
